!> シミュレーションで使用する物理定数を定義する。
module bem_constants
  use bem_kinds, only: dp
  implicit none
  real(dp), parameter :: k_coulomb = 8.9875517923d9
  real(dp), parameter :: k_boltzmann = 1.380649d-23
end module bem_constants
